@@ -1,21 +1,32 @@
 # coding: utf8
-""" Cog code generation tool.
+""" Cog content generation tool.
     http://nedbatchelder.com/code/cog
 
-    Copyright 2004-2016, Ned Batchelder.
+    Copyright 2004-2019, Ned Batchelder.
 """
 
 from __future__ import absolute_import, print_function
 
-import copy, getopt, hashlib, imp, os, re, shlex, sys, traceback
+import copy
+import getopt
+import glob
+import hashlib
+import imp
+import linecache
+import os
+import re
+import shlex
+import sys
+import traceback
+
 from .backward import PY3, StringIO, string_types, to_bytes
 
-__all__ = ['Cog', 'CogUsageError']
+__all__ = ['Cog', 'CogUsageError', 'main']
 
-__version__ = '2.5.1'   # History at the end of the file.
+__version__ = '3.0.0'
 
 usage = """\
-cog - generate code with inlined Python code.
+cog - generate content with inlined Python code.
 
 cog [OPTIONS] [INFILE | @FILELIST] ...
 
@@ -31,6 +42,8 @@ OPTIONS:
     -I PATH     Add PATH to the list of directories for data files and modules.
     -n ENCODING Use ENCODING when reading and writing files.
     -o OUTNAME  Write the output to OUTNAME.
+    -p PROLOGUE Prepend the generator source with PROLOGUE. Useful to insert an
+                import line. Example: -p "import math"
     -r          Replace the input file with the output.
     -s STRING   Suffix all generated output lines with STRING.
     -U          Write the output with Unix newlines (only LF line-endings).
@@ -64,17 +77,23 @@ class CogError(Exception):
 class CogUsageError(CogError):
     """ An error in usage of command-line arguments in cog.
     """
-    pass    #pragma: no cover
+    pass
 
 class CogInternalError(CogError):
     """ An error in the coding of Cog. Should never happen.
     """
-    pass    #pragma: no cover
+    pass
 
 class CogGeneratedError(CogError):
     """ An error raised by a user's cog generator.
     """
-    pass    #pragma: no cover
+    pass
+
+class CogUserException(CogError):
+    """ An exception caught when running a user's cog generator.
+        The argument is the traceback message to print.
+    """
+    pass
 
 class Redirectable:
     """ An object with its own stdout and stderr files.
@@ -101,10 +120,11 @@ class Redirectable:
 class CogGenerator(Redirectable):
     """ A generator pulled from a source file.
     """
-    def __init__(self):
+    def __init__(self, options=None):
         Redirectable.__init__(self)
         self.markers = []
         self.lines = []
+        self.options = options or CogOptions()
 
     def parseMarker(self, l):
         self.markers.append(l)
@@ -125,7 +145,7 @@ class CogGenerator(Redirectable):
 
         return reindentBlock(self.lines, '')
 
-    def evaluate(self, cog, globals, fname='cog generator'):
+    def evaluate(self, cog, globals, fname):
         # figure out the right whitespace prefix for the output
         prefOut = whitePrefix(self.markers)
 
@@ -133,9 +153,10 @@ class CogGenerator(Redirectable):
         if not intext:
             return ''
 
-        # In Python 2.2, the last line has to end in a newline.
-        intext = "import cog\n" + intext + "\n"
-        code = compile(intext, str(fname), 'exec')
+        prologue = "import " + cog.cogmodulename + " as cog\n"
+        if self.options.sPrologue:
+            prologue += self.options.sPrologue + '\n'
+        code = compile(prologue + intext, str(fname), 'exec')
 
         # Make sure the "cog" module has our state.
         cog.cogmodule.msg = self.msg
@@ -144,7 +165,17 @@ class CogGenerator(Redirectable):
         cog.cogmodule.error = self.error
 
         self.outstring = ''
-        eval(code, globals)
+        try:
+            eval(code, globals)
+        except CogError:
+            raise
+        except:
+            typ, err, tb = sys.exc_info()
+            frames = (tuple(fr) for fr in traceback.extract_tb(tb.tb_next))
+            frames = find_cog_source(frames, prologue)
+            msg = "".join(traceback.format_list(frames))
+            msg += "{}: {}".format(typ.__name__, err)
+            raise CogUserException(msg)
 
         # We need to make sure that the last line in the output
         # ends with a newline, or it will be joined to the
@@ -227,6 +258,7 @@ class CogOptions:
         self.sEndOutput = '[[[end]]]'
         self.sEncoding = "utf-8"
         self.verbosity = 2
+        self.sPrologue = ''
 
     def __eq__(self, other):
         """ Comparison operator for tests to use.
@@ -249,7 +281,7 @@ class CogOptions:
         try:
             opts, self.args = getopt.getopt(
                 argv,
-                'cdD:eI:n:o:rs:Uvw:xz',
+                'cdD:eI:n:o:rs:p:Uvw:xz',
                 [
                     'markers=',
                     'verbosity=',
@@ -281,6 +313,8 @@ class CogOptions:
                 self.bReplace = True
             elif o == '-s':
                 self.sSuffix = a
+            elif o == '-p':
+                self.sPrologue = a
             elif o == '-U':
                 self.bNewlines = True
             elif o == '-v':
@@ -325,6 +359,7 @@ class Cog(Redirectable):
         Redirectable.__init__(self)
         self.options = CogOptions()
         self._fixEndOutputPatterns()
+        self.cogmodulename = "cog"
         self.installCogModule()
 
     def _fixEndOutputPatterns(self):
@@ -350,7 +385,6 @@ class Cog(Redirectable):
         """
         self.cogmodule = imp.new_module('cog')
         self.cogmodule.path = []
-        sys.modules['cog'] = self.cogmodule
 
     def openOutputFile(self, fname):
         """ Open an output file, taking all the details into account.
@@ -404,6 +438,10 @@ class Cog(Redirectable):
 
             self.cogmodule.inFile = sFileIn
             self.cogmodule.outFile = sFileOut
+            self.cogmodulename = 'cog_' + hashlib.md5(sFileOut.encode()).hexdigest()
+            sys.modules[self.cogmodulename] = self.cogmodule
+            # if "import cog" explicitly done in code by user, note threading will cause clashes. 
+            sys.modules['cog'] = self.cogmodule
 
             # The globals dict we'll use for this file.
             if globals is None:
@@ -431,7 +469,7 @@ class Cog(Redirectable):
                     fOut.write(l)
 
                 # l is the begin spec
-                gen = CogGenerator()
+                gen = CogGenerator(options=self.options)
                 gen.setOutput(stdout=self.stdout)
                 gen.parseMarker(l)
                 firstLineNum = fIn.linenumber()
@@ -503,7 +541,7 @@ class Cog(Redirectable):
                 # supposed to generate code.
                 hasher = hashlib.md5()
                 if not self.options.bNoGenerate:
-                    sFile = "%s+%d" % (sFileIn, firstLineNum)
+                    sFile = "<cog %s:%d>" % (sFileIn, firstLineNum)
                     sGen = gen.evaluate(cog=self, globals=globals, fname=sFile)
                     sGen = self.suffixLines(sGen)
                     hasher.update(to_bytes(sGen))
@@ -546,7 +584,7 @@ class Cog(Redirectable):
 
 
     # A regex for non-empty lines, used by suffixLines.
-    reNonEmptyLines = re.compile("^\s*\S+.*$", re.MULTILINE)
+    reNonEmptyLines = re.compile(r"^\s*\S+.*$", re.MULTILINE)
 
     def suffixLines(self, text):
         """ Add suffixes to the lines in text, if our options desire it.
@@ -645,6 +683,14 @@ class Cog(Redirectable):
         finally:
             self.restoreIncludePath()
 
+    def processWildcards(self, sFile):
+        files = glob.glob(sFile)
+        if files:
+            for sMatchingFile in files:
+                self.processOneFile(sMatchingFile)
+        else:
+            self.processOneFile(sFile)
+
     def processFileList(self, sFileList):
         """ Process the files in a file list.
         """
@@ -676,7 +722,7 @@ class Cog(Redirectable):
                 raise CogUsageError("Can't use -o with @file")
             self.processFileList(args[0][1:])
         else:
-            self.processOneFile(args[0])
+            self.processWildcards(args[0])
 
         self.options = saved_options
 
@@ -720,62 +766,43 @@ class Cog(Redirectable):
         except CogGeneratedError as err:
             self.prerr("Error: %s" % err)
             return 3
+        except CogUserException as err:
+            self.prerr("Traceback (most recent call last):")
+            self.prerr(err.args[0])
+            return 4
         except CogError as err:
             self.prerr(err)
             return 1
-        except:
-            traceback.print_exc(None, self.stderr)
-            return 1
 
-# History:
-# 20040210: First public version.
-# 20040220: Text preceding the start and end marker are removed from Python lines.
-#           -v option on the command line shows the version.
-# 20040311: Make sure the last line of output is properly ended with a newline.
-# 20040605: Fixed some blank line handling in cog.
-#           Fixed problems with assigning to xml elements in handyxml.
-# 20040621: Changed all line-ends to LF from CRLF.
-# 20041002: Refactor some option handling to simplify unittesting the options.
-# 20041118: cog.out and cog.outl have optional string arguments.
-# 20041119: File names weren't being properly passed around for warnings, etc.
-# 20041122: Added cog.firstLineNum: a property with the line number of the [[[cog line.
-#           Added cog.inFile and cog.outFile: the names of the input and output file.
-# 20041218: Single-line cog generators, with start marker and end marker on
-#           the same line.
-# 20041230: Keep a single globals dict for all the code fragments in a single
-#           file so they can share state.
-# 20050206: Added the -x switch to remove all generated output.
-# 20050218: Now code can be on the marker lines as well.
-# 20050219: Added -c switch to checksum the output so that edits can be
-#           detected before they are obliterated.
-# 20050521: Added cog.error, contributed by Alexander Belchenko.
-# 20050720: Added code deletion and settable globals contributed by Blake Winton.
-# 20050724: Many tweaks to improve code coverage.
-# 20050726: Error messages are now printed with no traceback.
-#           Code can no longer appear on the marker lines,
-#               except for single-line style.
-#           -z allows omission of the [[[end]]] marker, and it will be assumed
-#               at the end of the file.
-# 20050729: Refactor option parsing into a separate class, in preparation for
-#               future features.
-# 20050805: The cogmodule.path wasn't being properly maintained.
-# 20050808: Added the -D option to define a global value.
-# 20050810: The %s in the -w command is dealt with more robustly.
-#           Added the -s option to suffix output lines with a marker.
-# 20050817: Now @files can have arguments on each line to change the cog's
-#               behavior for that line.
-# 20051006: Version 2.0
-# 20080521: -U options lets you create Unix newlines on Windows.  Thanks,
-#               Alexander Belchenko.
-# 20080522: It's now ok to have -d with output to stdout, and now we validate
-#               the args after each line of an @file.
-# 20090520: Use hashlib where it's available, to avoid a warning.
-#           Use the builtin compile() instead of compiler, for Jython.
-#           Explicitly close files we opened, Jython likes this.
-# 20120205: Port to Python 3.  Lowest supported version is 2.6.
-# 20150104: --markers option added by Doug Hellmann.
-# 20150104: -n ENCODING option added by Petr Gladkiy.
-# 20150107: Added --verbose to control what files get listed.
-# 20150111: Version 2.4
-# 20160213: v2.5: -o makes needed directories, thanks Jean-François Giraud.
-# 20161019: Added a LICENSE.txt file.
+
+def find_cog_source(frame_summary, prologue):
+    """Find cog source lines in a frame summary list, for printing tracebacks.
+
+    Arguments:
+        frame_summary: a list of 4-item tuples, as returned by traceback.extract_tb.
+        prologue: the text of the code prologue.
+
+    Returns
+        A list of 4-item tuples, updated to correct the cog entries.
+
+    """
+    prolines = prologue.splitlines()
+    for filename, lineno, funcname, source in frame_summary:
+        if not source:
+            m = re.search(r"^<cog ([^:]+):(\d+)>$", filename)
+            if m:
+                if lineno <= len(prolines):
+                    filename = '<prologue>'
+                    source = prolines[lineno-1]
+                    lineno -= 1     # Because "import cog" is the first line in the prologue
+                else:
+                    filename, coglineno = m.groups()
+                    coglineno = int(coglineno)
+                    lineno += coglineno - len(prolines)
+                    source = linecache.getline(filename, lineno).strip()
+        yield filename, lineno, funcname, source
+
+
+def main():
+    """Main function for entry_points to use."""
+    return Cog().main(sys.argv)
